#ifndef MULTILEVEL_PATCH_PRECONDITIONER_HH
#define MULTILEVEL_PATCH_PRECONDITIONER_HH

#include <string>

#include <dune/common/timer.hh>
#include <dune/common/fvector.hh>
#include <dune/common/bitsetvector.hh>

#include <dune/solvers/iterationsteps/lineariterationstep.hh>

#include <dune/faultnetworks/preconditioners/levelglobalpreconditioner.hh>
#include <dune/faultnetworks/preconditioners/levelpatchpreconditioner.hh>
#include <dune/faultnetworks/levelinterfacenetwork.hh>
#include <dune/faultnetworks/interfacenetwork.hh>
#include <dune/faultnetworks/dgmgtransfer.hh>

#include <dune/faultnetworks/utils/debugutils.hh>

template <class BasisType, class LocalOperatorAssembler, class LocalInterfaceAssembler, class MatrixType, class VectorType>
class MultilevelPatchPreconditioner : public LinearIterationStep<MatrixType, VectorType> {

public:
    enum Mode {additive, multiplicative};
    typedef LevelPatchPreconditioner<BasisType, LocalOperatorAssembler, LocalInterfaceAssembler, MatrixType, VectorType> LevelPatchPreconditionerType;
    typedef LevelGlobalPreconditioner<BasisType, LocalOperatorAssembler, LocalInterfaceAssembler, MatrixType, VectorType> LevelGlobalPreconditionerType;

private:
    typedef typename BasisType::GridView GridView;
    typedef typename GridView::Grid GridType;


    typedef typename Dune::BitSetVector<1> BitVector;

    const InterfaceNetwork<GridType>& interfaceNetwork_;
    const BitVector& activeLevels_;
    const std::vector<std::shared_ptr<LocalOperatorAssembler>>& localAssemblers_;
    const std::vector<std::vector<std::shared_ptr<LocalInterfaceAssembler>>>& localInterfaceAssemblers_;
    const Mode mode_;

    int itCounter_;

    std::shared_ptr<LevelGlobalPreconditionerType> coarseGlobalPreconditioner_;
    std::vector<std::shared_ptr<LevelPatchPreconditionerType>> levelPatchPreconditioners_;
    std::vector<VectorType> levelX_;
    std::vector<VectorType> levelRhs_;

    std::shared_ptr<LevelInterfaceNetwork<GridView>> allFaultLevelInterfaceNetwork_;
    std::shared_ptr<BasisType> allFaultBasis_;

    std::vector<std::shared_ptr<DGMGTransfer<BasisType>>> mgTransfer_;

public:
    MultilevelPatchPreconditioner(const InterfaceNetwork<GridType>& interfaceNetwork,
                                  const BitVector& activeLevels,
                                  const std::vector<std::shared_ptr<LocalOperatorAssembler>>& localAssemblers,
                                  const std::vector<std::vector<std::shared_ptr<LocalInterfaceAssembler>>>& localInterfaceAssemblers,
                                  const Mode mode = MultilevelPatchPreconditioner::Mode::additive) :
          interfaceNetwork_(interfaceNetwork),
          activeLevels_(activeLevels),
          localAssemblers_(localAssemblers),
          localInterfaceAssemblers_(localInterfaceAssemblers),
          mode_(mode)
    {

        if (activeLevels_.size() > (size_t) interfaceNetwork.maxLevel() +1)
            DUNE_THROW(Dune::Exception, "MultilevelFaultPreconditioner: too many active levels; preconditioner supports at most (grid.maxLevel + 1) levels!");

        assert(activeLevels.size() == localAssemblers_.size() && activeLevels.size() == localInterfaceAssemblers_.size());

        // init level fault preconditioners and multigrid transfer
        levelPatchPreconditioners_.resize(0);
        mgTransfer_.resize(0);

        int maxLevel = -1;
        for (size_t i=0; i<activeLevels_.size(); i++) {
            if (activeLevels_[i][0]) {
                // init global problem on coarsest level
                const LevelInterfaceNetwork<GridView>& levelNetwork = interfaceNetwork.levelInterfaceNetwork(i);
                coarseGlobalPreconditioner_ = std::make_shared<LevelGlobalPreconditionerType>(levelNetwork, *localAssemblers_[i], localInterfaceAssemblers_[i]);

                maxLevel = i;
                break;
            }
        }

        typename LevelPatchPreconditionerType::Mode levelMode = LevelPatchPreconditionerType::Mode::additive;
        if (mode_ != additive){
            levelMode = LevelPatchPreconditionerType::Mode::multiplicative;
        }

        for (size_t i=maxLevel+1; i<activeLevels_.size(); i++) {
            if (activeLevels_[i][0]) {
                // init local patch preconditioners on each level
                const LevelInterfaceNetwork<GridView>& levelNetwork = interfaceNetwork.levelInterfaceNetwork(i);

                //levelPatchPreconditioners_.push_back(std::make_shared<LevelGlobalPreconditionerType>(levelNetwork, *localAssemblers_[i], localInterfaceAssemblers_[i]));

                if (levelPatchPreconditioners_.size() == 0) {
                    levelPatchPreconditioners_.push_back(std::make_shared<LevelPatchPreconditionerType>(levelNetwork, coarseGlobalPreconditioner_->basis(), *localAssemblers_[i], localInterfaceAssemblers_[i], levelMode));
                } else {
                    //levelPatchPreconditioners_.push_back(std::make_shared<LevelPatchPreconditionerType>(levelNetwork, coarseGlobalPreconditioner_->basis(), *localAssemblers_[i], localInterfaceAssemblers_[i], levelMode));

                    levelPatchPreconditioners_.push_back(std::make_shared<LevelPatchPreconditionerType>(levelNetwork, levelPatchPreconditioners_.back()->basis(), *localAssemblers_[i], localInterfaceAssemblers_[i], levelMode));
                }

                maxLevel = i;
            }
        }

        levelX_.resize(levelPatchPreconditioners_.size()+1);
        levelRhs_.resize(levelPatchPreconditioners_.size()+1);

        allFaultBasis_ = std::make_shared<BasisType>(interfaceNetwork_.levelInterfaceNetwork(maxLevel));

        // init multigrid transfer
        for (size_t i=0; i<levelPatchPreconditioners_.size(); i++) {
            mgTransfer_.push_back(std::make_shared<DGMGTransfer<BasisType>>(levelPatchPreconditioners_[i]->basis(), *allFaultBasis_));
        }
        mgTransfer_.push_back(std::make_shared<DGMGTransfer<BasisType>>(coarseGlobalPreconditioner_->basis(), *allFaultBasis_));

        itCounter_ = 0;
    }


   virtual void setProblem(const MatrixType& mat, VectorType& x, const VectorType& rhs) {
        this->x_ = &x;
        this->rhs_ = &rhs;
        this->mat_ = Dune::stackobject_to_shared_ptr(mat);

        for (size_t i=0; i<levelPatchPreconditioners_.size(); i++) {
            mgTransfer_[i]->restrict(x, levelX_[i]);
            mgTransfer_[i]->restrict(rhs, levelRhs_[i]);

            levelPatchPreconditioners_[i]->setProblem(mat, levelX_[i], levelRhs_[i]);
        }

        size_t j = levelPatchPreconditioners_.size();
        mgTransfer_[j]->restrict(x, levelX_[j]);
        mgTransfer_[j]->restrict(rhs, levelRhs_[j]);

        coarseGlobalPreconditioner_->setProblem(mat, levelX_[j], levelRhs_[j]);
    }


    virtual void iterate() {
        if (mode_ == additive)
            iterateAdd();
        else
            iterateMult();
    }

    void iterateAdd() {
        *(this->x_) = 0;	

        VectorType x;

        for (size_t i=0; i<levelPatchPreconditioners_.size(); i++) {
            levelPatchPreconditioners_[i]->iterate();
            const VectorType& it = levelPatchPreconditioners_[i]->getSol();

            mgTransfer_[i]->prolong(it, x);

            //writeToVTK(*allFaultBasis_, x, "/home/mi/podlesny/data/faultnetworks/sol/", "preconditioner_step_"+std::to_string(i));

            *(this->x_) += x;
        }

        coarseGlobalPreconditioner_->iterate();
        const VectorType& it = coarseGlobalPreconditioner_->getSol();

        mgTransfer_[levelPatchPreconditioners_.size()]->prolong(it, x);
        *(this->x_) += x;


        /*
            levelPatchPreconditioners_[itCounter_]->iterate();
            const VectorType& it = levelPatchPreconditioners_[itCounter_]->getSol();

            mgTransfer_[itCounter_]->prolong(it, x);

            //writeToVTK(*allFaultBasis_, x, "/home/mi/podlesny/data/faultnetworks/sol/", "preconditioner_step_"+std::to_string(i));

            *(this->x_) += x;

        coarseGlobalPreconditioner_->iterate();
        const VectorType& it2 = coarseGlobalPreconditioner_->getSol();

        mgTransfer_[levelPatchPreconditioners_.size()]->prolong(it2, x);
        *(this->x_) += x;

        itCounter_++;
        itCounter_ = itCounter_ % (levelPatchPreconditioners_.size());

        if (itCounter_==0)
            std::cout << "Multilevel pass complete!" << std::endl;

        */
    }


    void iterateMult() {
        *(this->x_) = 0;

        VectorType x;

        for (int i=(levelPatchPreconditioners_.size()-1); i>=0; i--) {
            VectorType updatedRhs(*(this->rhs_));
            this->mat_->mmv(*(this->x_), updatedRhs);

            //print(updatedRhs, "globalRhs: ");

            mgTransfer_[i]->restrict(*(this->x_), levelX_[i]);
            mgTransfer_[i]->restrict(updatedRhs, levelRhs_[i]);

            //print(levelRhs_[i], "levelLocalCoarseRhs: ");
            //writeToVTK(levelPatchPreconditioners_[i]->basis(), levelRhs_[i], "/storage/mi/podlesny/data/faultnetworks/rhs/fine", "exactvertexdata_step_"+std::to_string(itCounter_));

            levelPatchPreconditioners_[i]->setProblem(*(this->mat_), levelX_[i], levelRhs_[i]);

            levelPatchPreconditioners_[i]->iterate();
            const VectorType& it = levelPatchPreconditioners_[i]->getSol();

            mgTransfer_[i]->prolong(it, x);

            //writeToVTK(*allFaultBasis_, x, "/storage/mi/podlesny/data/faultnetworks/fineIterates/multilevel", "exactvertexdata_step_"+std::to_string(itCounter_));


            *(this->x_) += x;
        }

        VectorType updatedCoarseRhs(*(this->rhs_));
        this->mat_->mmv(*(this->x_), updatedCoarseRhs);

        size_t j = levelPatchPreconditioners_.size();
        mgTransfer_[j]->restrict(*(this->x_), levelX_[j]);
        mgTransfer_[j]->restrict(updatedCoarseRhs, levelRhs_[j]);

        //print(levelRhs_[j], "localCoarseRhs: ");
        //writeToVTK(coarseGlobalPreconditioner_->basis(), levelRhs_[j], "/storage/mi/podlesny/data/faultnetworks/rhs/coarse", "exactvertexdata_step_"+std::to_string(itCounter_));

        coarseGlobalPreconditioner_->setProblem(*(this->mat_), levelX_[j], levelRhs_[j]);
        coarseGlobalPreconditioner_->iterate();
        const VectorType& it = coarseGlobalPreconditioner_->getSol();

        mgTransfer_[j]->prolong(it, x);

        //writeToVTK(*allFaultBasis_, x, "/storage/mi/podlesny/data/faultnetworks/coarseIterates/multilevel", "exactvertexdata_step_"+std::to_string(itCounter_));

        *(this->x_) += x;

        itCounter_++;
    }

    void build() {
        coarseGlobalPreconditioner_->build();

        for (size_t i=0; i<levelPatchPreconditioners_.size(); i++) {
            levelPatchPreconditioners_[i]->build();
        }
    }

    std::shared_ptr<LevelPatchPreconditionerType> levelPatchPreconditioner(const int level) {
        return levelPatchPreconditioners_[level];
    }

    const BasisType& basis() const {
        return *allFaultBasis_;
    }

    size_t size() const {
        return levelPatchPreconditioners_.size()+1;
    }
};

#endif

